# -*- coding: utf-8 -*-

from inventree import base
from inventree import part


class PurchaseOrder(base.InventreeObject):
    """ Class representing the PurchaseOrder database model """

    URL = 'order/po'
    FILTERS = [
        'status',
        'part',
        'supplier_part',
        'supplier',
    ]

    def __repr__(self):
        return "Purchase Order #{ref}".format(ref=self.pk)

    def getLineItems(self):
        """ Return the line items associated with this order """
        return PurchaseOrderLineItem.list(self._api, order=self.pk)


class PurchaseOrderLineItem(base.InventreeObject):
    """ Class representing the PurchaseOrderLineItem database model """

    URL = 'order/po-line/'
    FILTERS = [
        'order',
        'part',
    ]

    def __repr__(self):
        return "{n} x part #{part} for order #{ref}".format(
            n=self.quantity,
            part=self.part,
            ref=self.order
        )

    def getPart(self):
        return part.Part(self._api, self.part)

    def getOrder(self):
        return PurchaseOrder(self._api, self.order)


class SalesOrder(base.InventreeObject):
    """ Class respresenting the SalesOrder database model """

    URL = 'order/so'
    FILTERS = [
        'status',
        'part',
        'customer',
    ]

    def __repr__(self):
        return "Sales Order #{ref}".format(ref=self.pk)

    def getLineItems(self):
        """ Return the line items associated with this order """
        return SalesOrderLineItem.list(self._api, order=self.pk)


class SalesOrderLineItem(base.InventreeObject):
    """ Class representing the SalesOrderLineItem database model """

    URL = 'order/so-line/'
    FILTERS = [
        'order',
        'part'
    ]

    def __repr__(self):
        return "{n} x part #{part} for order #{ref}".format(
            n=self.quantity,
            part=self.part,
            ref=self.order,
        )

    def getPart(self):
        return part.Part(self._api, self.part)

    def getOrder(self):
        return SalesOrder(self._api, self.order)



# -*- coding: utf-8 -*-

import os
import logging

INVENTREE_PYTHON_VERSION = "0.0.7"


class InventreeObject():
    """ Base class for an InvenTree object """

    URL = ""
    FILTERS = []

    # Special filters
    FILTERS_EXTRA = ['search', 'order']

    def __init__(self, api, pk=None, data={}):
        """ Instantiate this InvenTree object.

        Args:
            pk - The ID (primary key) associated with this object on the server
            api - The request manager object
            data - JSON representation of the object
        """

        # If the pk is not explicitly provided,
        # extract it from the provided dataset
        if pk is None:
            pk = data['pk']

        self._url = "{url}/{pk}/".format(url=self.URL, pk=pk)
        self._api = api
        self._data = data

        # If the data are not populated, fetch from server
        if len(self._data) == 0:
            self.reload()

    @property
    def pk(self):
        """ Convenience method for accessing primary-key field """
        return self['pk']

    @classmethod
    def create(cls, api, data, **kwargs):
        """ Create a new database object in this class. """

        # Ensure the pk value is None so an existing object is not updated
        if 'pk' in data.keys():
            data.pop('pk')

        api.post(cls.URL, data)

    @classmethod
    def list(cls, api, **kwargs):
        """ Return a list of all items in this class on the database.

        Requires:

        URL - Base URL
        FILTERS - List of available query filter params
        """

        # Dict of query params to send to the API
        params = {}

        allowed = cls.FILTERS + cls.FILTERS_EXTRA

        for arg in allowed:
            if kwargs.get(arg, None):
                params[arg] = kwargs.pop(arg)

        response = api.get(cls.URL, params=params, **kwargs)

        if response is None:
            return None

        items = []

        for data in response:
            if 'pk' in data:
                items.append(cls(data=data, api=api))

        return items

    def delete(self):
        """ Delete this object from the database """
        if self._api:
            self._api.delete(self._url)

    def save(self):
        """ Save this object to the database """
        if self._api:
            self._api.put(self._url, self._data)

    def reload(self):
        """ Reload object data from the database """
        if self._api:
            data = self._api.get(self._url)
            if data is not None:
                self._data = data

    def __getattr__(self, name):
        if name in self._data.keys():
            return self._data[name]
        else:
            return super().__getattr__(name)

    def __getitem__(self, name):
        if name in self._data.keys():
            return self._data[name]
        else:
            raise KeyError("Key '{k}' does not exist in dataset".format(k=name))

    def __setitem__(self, name, value):
        if name in self._data.keys():
            self._data[name] = value
        else:
            raise KeyError("Key '{k}' does not exist in dataset".format(k=name))


class Attachment(InventreeObject):
    """ Class representing a file attachment object """

    @classmethod
    def upload(cls, api, filename, comment, **kwargs):
        """
        Upload a file attachment.
        Ref: https://2.python-requests.org/en/master/user/quickstart/#post-a-multipart-encoded-file
        """

        if not os.path.exists(filename):
            logging.error("File does not exist: '{f}'".format(f=filename))
            return

        f = os.path.basename(filename)

        data = kwargs

        # File comment must be provided
        data['comment'] = comment

        files = {
            'attachment': (f, open(filename, 'rb')),
        }

        # Send the file off to the server
        if api.post(cls.URL, data, files=files):
            logging.info("Uploaded attachment file: '{f}'".format(f=f))
        else:
            logging.warning("File upload failed")


class Currency(InventreeObject):
    """ Class representing the Currency database model """

    URL = 'common/currency'


class Parameter(InventreeObject):
    """class representing the Parameter database model """
    URL = 'part/parameter'
    FILTERS = ['part']

    def getunits(self):
        """ Get the dimension and units for this parameter """

        return [element for element
                in ParameterTemplate.list(self._api)
                if element['pk'] == self._data['template']]


class ParameterTemplate(InventreeObject):
    """ class representing the Parameter Template database model"""

    URL = 'part/parameter/template'



# -*- coding: utf-8 -*-

"""
The inventree_api module handles low level requests and authentication
with the InvenTree database server.
"""


import requests
from requests.auth import HTTPBasicAuth
import os
import json
import logging


class InvenTreeAPI(object):
    """ Basic class for performing Inventree API requests.

    GET - Fetch data from the server

    """

    def __init__(self, base_url, **kwargs):
        """ Initialize class with initial parameters

        Args:
            base_url - Base API URL
            
        kwargs:
            username - Login username
            password - Login password
            token - Authentication token (if provided, username/password are ignored)
            use_token_auth - Use token authentication? (default = True)
            verbose - Print extra debug messages (default = False)
        """

        if not base_url.endswith('/'):
            base_url += '/'

        # Server address *must* end with /api/
        if not base_url.endswith('/api/'):
            base_url = os.path.join(base_url, 'api')

        if not base_url.endswith('/'):
            base_url += '/'

        self.base_url = base_url

        logging.info("Connecting to server: " + str(self.base_url))

        self.username = kwargs.get('username', None)
        self.password = kwargs.get('password', None)
        self.token = kwargs.get('token', None)
        self.use_token_auth = kwargs.get('use_token_auth', True)
        self.verbose = kwargs.get('verbose', False)

        # Check if the server is there
        self.testServer()

        # Basic authentication
        self.auth = HTTPBasicAuth(self.username, self.password)
        
        if self.use_token_auth:
            if not self.token:
                self.requestToken()

    def clean_url(self, url):

        url = os.path.join(self.base_url, url)

        if not url.endswith('/'):
            url += '/'

        return url

    def testServer(self):
        """
        Check to see if the server is present.
        The InvenTree server provides a simple endpoint at /api/
        which contains some simple data (and does not require authentication)
        """

        self.server_details = None

        logging.info("Checking InvenTree server connection...")

        try:
            response = requests.get(self.base_url)
        except requests.exceptions.ConnectionError:
            logging.error("Server connection refused - check server address")
            return False

        if not response.status_code == 200:
            logging.error("Error code from server: {code} - {detail}".format(
                code=response.status_code,
                detail=response.text
            ))

            return False

        # Record server details
        self.server_details = json.loads(response.text)

        logging.info("InvenTree server details: " + str(response.text))

        # The details provided by the server should include some specific data:
        server_name = str(self.server_details.get('server', ''))

        if not server_name.lower() == 'inventree':
            logging.warning("Server returned strange response (expected 'InvenTree', found '{name}')".format(
                name=server_name
            ))

        return True

    def requestToken(self):
        """ Return authentication token from the server """

        if not self.username or not self.password:
            raise AttributeError('Supply username and password to request token')

        logging.info("Requesting auth token from server...")

        # Request an auth token from the server
        token_url = os.path.join(self.base_url, 'user/token/')
        
        reply = requests.get(token_url, auth=self.auth)

        data = json.loads(reply.text)

        if not reply.status_code == 200:
            logging.error("Error requesting token: {code} - {detail}".format(
                code=reply.status_code,
                detail=str(reply.text)
            ))
            return None

        if 'token' not in data.keys():
            logging.error("Token not returned by server: {detail}".format(
                detail=str(reply.text)
            ))
            return None

        self.token = json.loads(reply.text)['token']

        logging.info("Authentication token: " + self.token)

        return self.token

    def request(self, url, **kwargs):
        """ Perform a URL request to the Inventree API """

        # Remove leading slash
        if url.startswith('/'):
            url = url[1:]

        api_url = os.path.join(self.base_url, url)

        if not api_url.endswith('/'):
            api_url += '/'

        method = kwargs.get('method', 'get')

        params = kwargs.get('params', {})

        json = kwargs.get('json', {})

        headers = kwargs.get('headers', {})

        search_term = kwargs.get('search', None)

        if search_term is not None:
            params['search'] = search_term

        methods = {
            'GET': requests.get,
            'POST': requests.post,
            'PUT': requests.put,
            'DELETE': requests.delete,
        }

        if method.upper() not in methods.keys():
            logging.error("Unknown request method '{m}'".format(m=method))
            return None

        method = method.upper()

        if self.use_token_auth and self.token:
            headers['AUTHORIZATION'] = 'Token {t}'.format(t=self.token)
            auth = None
        else:
            auth = self.auth

        logging.debug("Sending Request:")
        logging.debug(" - URL:", method, api_url)
        logging.debug(" - auth:", auth)
        logging.debug(" - params:", params)
        logging.debug(" - headers:", headers)
        logging.debug(" - json:", json)

        try:
            response = methods[method](
                api_url,
                auth=auth,
                params=params,
                headers=headers,
                json=json
            )

        except requests.exceptions.ConnectionError:
            logging.error("Connection refused - '{url}'".format(url=api_url))
            return None

        if response is None:
            logging.error("Null response - {method} '{url}'".format(method=method, url=api_url))
            return None

        logging.info("Request: {method} {url} - {response}".format(method=method, url=api_url, response=response.status_code))

        # Detect invalid response codes
        # Anything 300+ is 'bad'
        if response.status_code >= 300:
            logging.warning("Bad response ({code}) - {method} '{url}' - {detail}".format(
                code=response.status_code, method=method, url=api_url,
                detail=str(response.text)
            ))

        # A delete request won't return JSON formatted data (ignore further checks)
        if method == 'DELETE':
            return response

        ctype = response.headers.get('content-type')

        if not ctype == 'application/json':
            logging.error("'Response content-type is not JSON - '{url}' - '{f}'".format(url=api_url, f=ctype))
            return None

        return response

    def delete(self, url, **kwargs):
        """ Perform a DELETE request. Used to remove a record in the database.

        """

        headers = {'content-type': 'application/json'}

        response = self.request(url, method='delete', headers=headers, **kwargs)

        if response is None:
            return False

        print(response.status_code, response.text)

    def post(self, url, data, files=None, **kwargs):
        """ Perform a POST request. Used to create a new record in the database.

        Args:
            url - API endpoint URL
            data - JSON data to create new object
            files - Dict of file attachments
        """

        url = self.clean_url(url)

        headers = {}

        if self.use_token_auth and self.token:
            headers['AUTHORIZATION'] = 'Token {t}'.format(t=self.token)
            auth = None
        else:
            auth = self.auth

        response = requests.post(url, data=data, headers=headers, auth=auth, files=files, **kwargs)

        if response is None:
            return False

        if response.status_code in [200, 201]:
            return True
        else:
            logging.error("POST request failed at '{url}' - {status}".format(url=url, status=response.status_code))
            logging.error(response.text)
            return False

    def put(self, url, data, **kwargs):
        """ Perform a PUT request. Used to update existing records in the database.

        Args:
            url - API endpoint URL
            data - JSON data to PUT
        """

        headers = {'content-type': 'application/json'}

        params = {
            'format': 'json',
        }

        response = self.request(url, json=data, method='put', headers=headers, params=params, **kwargs)

        if response is None:
            return None
        
        if response.status_code == 200:
            return True
        else:
            logging.error("PUT request failed at '{url}' - {status}".format(url=url, status=response.status_code))
            return False

    def get(self, url, **kwargs):
        """ Perform a GET request

        Args:
            url - API url

        kwargs:

        """

        response = self.request(url, method='get', **kwargs)

        # No response returned
        if response is None:
            return None

        try:
            data = json.loads(response.text)
        except json.decoder.JSONDecodeError:
            logging.error("Error decoding JSON response - '{url}'".format(url=url))
            return None

        return data



# -*- coding: utf-8 -*-

"""
A simple script to test connection to an InvenTree server.
User must provide address / username / password information.
"""

import argparse
import logging
import sys

from inventree.api import InvenTreeAPI
from inventree.base import INVENTREE_PYTHON_VERSION


def test_server(address, **kwargs):
    """
    Perform basic server checks.
    """

    if kwargs.get('debug', False):
        # Print all debug messages
        logging.basicConfig(level=logging.INFO)

    api = InvenTreeAPI(address, **kwargs)

    assert api.server_details is not None
    assert api.token is not None

    print("Server Details:", api.server_details)
    print("Token:", api.token)

    print("All checks passed OK...")


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="InvenTree server test")

    parser.add_argument("address", help="InvenTree server address")
    parser.add_argument("-u", "--username", help="Username", default=None)
    parser.add_argument("-p", "--password", help="Password", default=None)
    parser.add_argument("-t", "--token", help="Authentication token", default=None)

    args = parser.parse_args()

    print("InvenTree python version: " + INVENTREE_PYTHON_VERSION)

    # User must provide either USERNAME/PASSWORD or Token
    if args.token is None:
        if args.username is None:
            logging.error("Username must be provided")
            sys.exit(1)
        
        if args.password is None:
            logging.error("Password must be provided")
            sys.exit(1)

    test_server(
        args.address,
        username=args.username,
        password=args.password,
        token=args.token
    )



# -*- coding: utf-8 -*-

import re

from inventree import base
from inventree import stock
from inventree import company
from inventree import build


class PartCategory(base.InventreeObject):
    """ Class representing the PartCategory database model """

    URL = 'part/category'
    FILTERS = ['parent']

    def getParts(self):
        return Part.list(self._api, category=self.pk)

    def getChildCategories(self):
        return PartCategory.list(self._api, parent=self.pk)
    

class Part(base.InventreeObject):
    """ Class representing the Part database model """

    URL = 'part'
    FILTERS = [
        'category',
        'cascade',
        'has_stock',
        'low_stock',
        'is_template',
        'variant_of',
        'assemply',
        'component',
        'trackable',
        'salable',
        'active',
        'buildable',
        'purchaseable',
    ]

    def getTestTemplates(self):
        """ Return all test templates associated with this part """
        return PartTestTemplate.list(self._api, part=self.pk)

    def getSupplierParts(self):
        """ Return the supplier parts associated with this part """
        return company.SupplierPart.list(self._api, part=self.pk)

    def getBomItems(self):
        """ Return the items required to make this part """
        return BomItem.list(self._api, part=self.pk)

    def isUsedIn(self):
        """ Return a list of all the parts this part is used in """
        return BomItem.list(self._api, sub_part=self.pk)

    def getBuilds(self):
        """ Return the builds associated with this part """
        return build.Build.list(self._api, part=self.pk)

    def getStockItems(self):
        """ Return the stock items associated with this part """
        return stock.StockItem.list(self._api, part=self.pk)

    def getAttachments(self):
        """ Return attachments associated with this part """
        return PartAttachment.list(self._api, part=self.pk)


class PartAttachment(base.Attachment):
    """ Class representing a file attachment for a Part """

    URL = 'part/attachment'
    FILTERS = ['part']


class PartTestTemplate(base.InventreeObject):
    """ Class representing a test template for a Part """

    URL = 'part/test-template'
    FILTERS = ['part', 'required']

    @classmethod
    def generateTestKey(cls, test_name):
        """ Generate a 'key' for this test """

        key = test_name.strip().lower()
        key = key.replace(' ', '')

        # Remove any characters that cannot be used to represent a variable
        key = re.sub(r'[^a-zA-Z0-9]', '', key)

        return key

    def getTestKey(self):
        return PartTestTemplate.generateTestKey(self.test_name)
    

class BomItem(base.InventreeObject):
    """ Class representing the BomItem database model """

    URL = 'bom'
    FILTERS = ['part', 'sub_part']



# -*- coding: utf-8 -*-

from inventree import base


class Company(base.InventreeObject):
    """ Class representing the Company database model """

    URL = 'company'
    FILTERS = ['is_supplier', 'is_customer']

    def getSuppliedParts(self):
        return SupplierPart.list(self._api, supplier=self.pk)

    def getManufacturedParts(self):
        return SupplierPart.list(self._api, manufacturer=self.pk)


class SupplierPart(base.InventreeObject):
    """ Class representing the SupplierPart database model """

    URL = 'company/part'
    FILTERS = ['part', 'supplier', 'manufacturer']

    def getPriceBreaks(self):
        """ Get a list of price break objects for this SupplierPart """

        return SupplierPriceBreak.list(self._api, part=self.pk)


class SupplierPriceBreak(base.InventreeObject):
    """ Class representing the SupplierPriceBreak database model """

    URL = 'company/price-break/'
    FILTERS = ['part', 'currency']



# -*- coding: utf-8 -*-

import os
import logging

from inventree import base
from inventree import part


class StockLocation(base.InventreeObject):
    """ Class representing the StockLocation database model """

    URL = 'stock/location'
    FILTERS = ['parent']

    def getStockItems(self):
        return StockItem.list(self._api, location=self.pk)


class StockItem(base.InventreeObject):
    """ Class representing the StockItem database model.
    
    Stock items can be filtered by:
    
    - location: Where the stock item is stored
    - category: The category of the part this stock item points to
    - supplier: Who supplied this stock item
    - part: The part referenced by this stock item
    - supplier_part: Matching SupplierPart object
    """

    URL = 'stock'
    FILTERS = [
        'location',
        'category',
        'cascade',
        'supplier',
        'part',
        'IPN',
        'supplier_part'
        'build',
        'build_order',
        'belongs_to',
        'sales_order',
        'customer',
        'serialized',
        'serial_number',
        'allocated',
        'active',
        'ancestor',
        'status',
        'company',
        'supplier',
        'manufacturer'
        'serialized',
        'serial',
    ]

    def getPart(self):
        """ Return the base Part object associated with this StockItem """
        return part.Part(self._api, self.part)

    def getAttachments(self):
        """ Return all file attachments for this StockItem """

        return StockItemAttachment.list(
            self._api,
            stock_item=self.pk
        )

    def uploadAttachment(self, filename, comment, **kwargs):
        """ Upload a file attachment against this StockItem """

        kwargs['stock_item'] = self.pk

        StockItemAttachment.upload(self._api, filename, comment, **kwargs)

    def getTestResults(self, **kwargs):
        """ Return all the test results associated with this StockItem """

        return StockItemTestResult.list(self._api, stock_item=self.pk)

    def uploadTestResult(self, test_name, test_result, **kwargs):
        """ Upload a test result against this StockItem """

        StockItemTestResult.upload_result(self._api, self.pk, test_name, test_result, **kwargs)


class StockItemAttachment(base.Attachment):
    """ Class representing a file attachment for a StockItem """

    URL = 'stock/attachment'
    FILTERS = ['stock_item']


class StockItemTracking(base.InventreeObject):
    """ Class representing a StockItem tracking object """

    URL = 'stock/track'
    FILTERS = ['item', 'user']


class StockItemTestResult(base.InventreeObject):
    """ Class representing a StockItemTestResult object """

    URL = 'stock/test'
    FILTERS = ['stock_item', 'test', 'result', 'value', 'user']

    def getTestKey(self):
        return part.PartTestTemplate.generateTestKey(self.test)

    @classmethod
    def upload_result(cls, api, stock_item, test, result, **kwargs):
        """
        Upload a test result.

        args:
            api: Authenticated InvenTree API object
            stock_item: pk of the StockItem object to upload the test result against
            test: Name of the test (string)
            result: Test result (boolean)

        kwargs:
            attachment: Optionally attach a file to the test result
            notes: Add extra notes
            value: Add a "value" to the test (e.g. an actual measurement made during the test)
        """

        attachment = kwargs.get('attachment', None)

        files = {}

        if attachment:
            if os.path.exists(attachment):
                f = os.path.basename(attachment)
                files['attachment'] = (f, open(attachment, 'rb'))
            else:
                logging.error("File does not exist: '{f}'".format(f=attachment))

        notes = kwargs.get('notes', '')
        value = kwargs.get('value', '')

        data = {
            'stock_item': stock_item,
            'test': test,
            'result': result,
            'notes': notes,
            'value': value,
        }

        # Send the data to the serever
        if api.post(cls.URL, data, files=files):
            logging.info("Uploaded test result: '{test}'".format(test=test))
        else:
            logging.warning("Test upload failed")
